/*

	process.c  -  routines voor manipulating processes including
	              process switching, adding etc.

    Copyright (C) 2007  J. op den Brouw <jesse@opdenbrouw.nl>

	This file is part of jrtos version 0.2

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include <avr/interrupt.h>
#include <avr/io.h>
#include <stdint.h>

#include "process.h"

/* The processes ;-) */
process_descriptor context[MAX_PROCESS];
/* Pointer to currect task context, used for fast access to the   */
/* current process, and also used in the assembler context save   */
/* and restore routines.                                          */
static volatile process_descriptor * volatile ptr = &context[0];
/* The number of the currect task, an index in the process table. */
static volatile uint8_t p=0;
/* Add pointer, used for indexing the process table when adding   */
/* new processes.                                                 */
static volatile uint8_t add_pointer=1;
static volatile switch_info_type switch_info;


/* The context switcher, it's really an interrupt handler with    */
/* special attributes. It first pushes the current context on     */
/* stack, updates the number of ticks ran, sets some info,        */
/* then picks a new task (simple round robin) and finally pops    */
/* the new task's context from its stack.                         */
void TIMER1_COMPA_vect( void )
{
	/* Note; NEVER use local variables, because they're pushed    */
	/* on stack.                                                  */
	/* Push current context.                                      */
	/* Idea is politely stolen from FreeRTOS.                     */
	/* This is really a macro, not a function call.               */
	save_context();

	/* How many ticks? */
	ptr->ticks+=TICKS_PER_SLICE-TICKS_IN_SAVE_CONTEXT;

	/* Record the fact that we are switched */
	ptr->switches++;

	/* Context switch info */
	switch_info.isr++;
	switch_info.ticks+=280;

	/* Context saved by ISR */
	ptr->flags=(ptr->flags&~BY_MASK)|BY_ISR;

	/* Select a new task to run */
	do {
		p++;
		if (p==MAX_PROCESS) { p = 0; }
	} while (context[p].state != ACTIVE);
	/* bug: if no process is active, this hangs! */

	/* Set pointer for asm routines */
	ptr = &context[p];

	/* Loads context of selected task. Idea is politely stolen from FreeRTOS */
	/* Restore_context() is really a macro, not a function call.             */
	/* First check if we were switched by sem_yield, return with a ret so    */
	/* interrupts stay disabled. If switched by an ISR or yield, return with */
	/* a reti which enables interrupts.                                      */
	restore_context();
}


/* Voluntarily give up my slice */
void yield(void) {
	/* Note; NEVER use local variables, because they're pushed */
	/* on stack */
	/* Push current context */
	/* This is politely stolen from FreeRTOS */
	save_context();

	/* How many ticks? */
	/*context[p].ticks+=(uint32_t)TCNT1;*/
	ptr->ticks+=(uint32_t)TCNT1-TICKS_IN_SAVE_CONTEXT;

	/* Record the fact that we are switched */
	ptr->switches++;

	/* Reset time slice, give it a little because we souped up some time */
	/*TCNT1=2;*/
	/* Context switch info */
	switch_info.yield++;
	switch_info.ticks+=280;


	/* Context saved by yield */
	/*context[p].flags|=BY_YIELD;*/
	ptr->flags=(ptr->flags&~BY_MASK)|BY_YIELD;

	/* Select a new task to run */
	do {
		p++;
		if (p==MAX_PROCESS) { p = 0; }
	} while (context[p].state != ACTIVE);
	/* bug: if no process is active, this hangs! */
	/* Set pointer for asm routines */
	ptr = &context[p];

	/* Loads context of selected task. This is politely stolen from FreeRTOS */
	/* Restore_context() is really a macro, not a function call.             */
	/* First check if we were switched by sem_yield, return with a ret so    */
	/* interrupts stay disabled. If switched by an ISR or yield, return with */
	/* a reti which enables interrupts.                                      */
	restore_context();
}


/* Voluntarily give up my slice, semaphore version */
void sem_yield(void) {
	/* Note; NEVER use local variables, because they're pushed */
	/* on stack */
	/* Push current context */
	/* This is politely stolen from FreeRTOS */
	save_context();

	/* How many ticks? */
	ptr->ticks+=(uint32_t)TCNT1-TICKS_IN_SAVE_CONTEXT;

	/* Record the fact that we are switched */
	ptr->switches++;

	/* Context switch info */
	switch_info.sem_yield++;
	switch_info.ticks+=280;

	/* Context saved by semaphore yield */
	ptr->flags=(ptr->flags&~BY_MASK)|BY_SEMYIELD;

	/* Select a new task to run */
	do {
		p++;
		if (p==MAX_PROCESS) { p = 0; }
	} while (context[p].state != ACTIVE);
	/* bug: if no process is active, this hangs! */

	/* Set pointer for asm routines */
	ptr = &context[p];

	/* Loads context of selected task. This is politely stolen from FreeRTOS */
	/* Restore_context() is really a macro, not a function call.             */
	/* First check if we were switched by sem_yield, return with a ret so    */
	/* interrupts stay disabled. If switched by an ISR or yield, return with */
	/* a reti which enables interrupts.                                      */
	restore_context();
}

/* Initialize the contexts. It does five things: puts the lost task catcher  */
/* address on stack, so tasks that return get handled well; sets stack       */
/* pointer to the right address so that on the task's first run, the         */
/* registers are loaded (cleared) from stack; sets the Y-pointer (r28 and    */
/* r29) to the top of the stack and finally sets the task state to empty and */
/* lost flag to alive.                                                       */
void clear_process_list(void) {
	uint8_t i;
	for (i=0; i<MAX_PROCESS; i++) {
		/* Put the return address of the lost proc catcher on stack */
		context[i].stack[STACK_LEN-1] = (uint8_t)  ((uint16_t)catch_lost_proc & 0xff);
		context[i].stack[STACK_LEN-2] = (uint8_t)  (((uint16_t)catch_lost_proc >> 8) & 0xff);
		/* Set the stack pointer, see next lines for how to calculate */
		/* STACK_LEN(64) - top-of-stack(1) - catch address(2) - call address (2) - regs(32) .... */
		/* ...  - post dec stack(1) */
		context[i].spl = (uint8_t)  ( (uint16_t)(context[i].stack) +STACK_LEN-1-2-2-32-1) & 0xff;
		context[i].sph = (uint8_t) (( (uint16_t)(context[i].stack) +STACK_LEN-1-2-2-32-1)>>8) & 0xff;
		/* Now set up the Y-pointer to free space on stack, this is for local variables */
		context[i].stack[STACK_LEN-9] =  ( (uint16_t)(context[i].stack) +STACK_LEN-3) & 0xff;
		context[i].stack[STACK_LEN-8] = (( (uint16_t)(context[i].stack) +STACK_LEN-3)>>8) & 0xff;
		/* Rest of the info. */
		context[i].flags = NOTHING;
		context[i].state=EMPTY;
		context[i].ticks = 0UL;
		context[i].switches = 0U;
	}
	/* Set pointer to 1, this skipping no. 0. */
	add_pointer=1;
	/* FIXME: the main holds context on task 0, but it doesn't use the stack */
	context[0].state=ACTIVE;
}

/* Load the contexts. It does two things: puts start adress of task                */
/* on stack so tasks get started anyway; sets stackpointer to the right address    */
/* so that on the task's first run, the registers are loaded (cleared) from stack; */
/* and finally sets the task state to active.                                      */
void add_process(void *pfunc) {
	context[add_pointer].stack[STACK_LEN-3] = (uint8_t)  ((uint16_t)pfunc & 0xff);
	context[add_pointer].stack[STACK_LEN-4] = (uint8_t)  (((uint16_t)pfunc >> 8) & 0xff);
	context[add_pointer].state=ACTIVE;
	add_pointer++;
}


static volatile uint8_t retaddrl, retaddrh;
void add_process_main(void) {

	/* Get return address */
	asm volatile ("pop r1");
	asm volatile ("sts retaddrh,r1");
	asm volatile ("pop r1");
	asm volatile ("sts retaddrl,r1");
	asm volatile ("clr r1");

	/* Place return address on process 0's stack */
	context[0].stack[STACK_LEN-3] = retaddrl;
	context[0].stack[STACK_LEN-4] = retaddrh;

	/* Calculate new stack pointer value */
	retaddrl  = (uint8_t)  ( (uint16_t)(context[0].stack) +STACK_LEN-4-1) & 0xff;
	retaddrh  = (uint8_t) (( (uint16_t)(context[0].stack) +STACK_LEN-4-1)>>8) & 0xff;

	/* Set the current stack pointer. */
	asm volatile ("lds r1,retaddrl");
	asm volatile ("out __SP_L__, r1");
	asm volatile ("lds r1,retaddrh");
	asm volatile ("out __SP_H__, r1");
	asm volatile ("clr r1");
}

/* This sets up the timer for pre-emptive context switches */
void setup_timer(void) {
	TIMSK  = (1<<4);
	TIFR   = (1<<4);
	TCCR1A = 0x00;	/* 0000.0000 */
	OCR1A  = (uint16_t) TICKS_PER_SLICE;
	TCCR1B = 0x08+(uint8_t)PRESCALER_BITMAP;  /* 0000.1xxx */
}


/* This routine is called when a concurrent task executes */
/* a return instruction. It cannot return to anything,    */
/* because it was not really called. What we do is flag   */
/* the task as stopped, and yield this function. The code */
/* for setting the task to lost scrambles the register    */
/* contents, but the task is ended anyway.                */
void catch_lost_proc(void) {
	cli();
	ptr->state=STOPPED;
	/* Picking an new process only is faster, but now the */
	/* context is also saved. Note that we never return   */
	/* to this routine.                                   */
	yield();
}

