/*

	process.h  -  header file to process.c

    Copyright (C) 2007  J. op den Brouw <jesse@opdenbrouw.nl>

	This file is part of jrtos version 0.2

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/


#ifndef _PROCESS_H_
#define _PROCESS_H_

#include <avr/interrupt.h>
#include <avr/io.h>
#include <stdint.h>

/* Check for device used, only the 16 and 32 are supported for now. */
#if defined (__AVR_ATmega16__)
#define RAMSIZE 1024
#elif defined (__AVR_ATmega32__)
#define RAMSIZE 2048
#else
#error AVR CPU not supported!!!
#endif

/* Defines on timer and time slice */
/* The F_CPU macro must be defined, otherwise compilation will fail. */
/* The TIME_SLICE macro holds the time slice time in MILLIseconds.   */
/* The PRESCALER macro holds one of 1, 8, 64, 256 or 1024.           */
#ifndef F_CPU
#error  F_CPU not defined!
#define F_CPU 3686400UL
#endif
#define TIME_SLICE 10UL
#define PRESCALER  64UL

/* Some defines on processes */
#define MAX_PROCESS 8
#define STACK_LEN 96
/* Process states */
#define EMPTY 0
#define ACTIVE 1
#define STOPPED 2
/* Process flags */
#define NOTHING 0
#define BY_ISR 1
#define BY_YIELD 2
#define BY_SEMYIELD 4
#define BY_FREE 8
#define BY_MASK 0x0f


/* Context storage */
typedef struct {
	/* The stackpointer must be the first one!!! */
	/* The flags must be number three!!!!!!!!!!! */
	uint8_t spl;				/* Stackpointer low */
	uint8_t sph;				/* Stackpointer high */
	uint8_t flags;				/* Some flags */
	uint8_t state;              /* The task's state, one of empty, active, stopped */
	uint32_t ticks;				/* Number of TC1 ticks ran */
	uint16_t switches;			/* Times the process is switched */
	uint8_t stack[STACK_LEN];	/* The task's stack */
} process_descriptor;

typedef struct {
	uint16_t isr;
	uint16_t yield;
	uint16_t sem_yield;
	uint32_t ticks;
} switch_info_type;

/* Calculate maximum numbers of processes when not defined. Subtract */
/* a few bytes for global storage of variables.                      */
#ifndef MAX_PROCESS
#define MAX_PROCESS ((RAMSIZE-200)/(sizeof(process_descriptor)))
#endif

/* Calculate the number of TC1 ticks per time slice */
#define TICKS_PER_SLICE  ((F_CPU*TIME_SLICE)/(PRESCALER*1000UL))

/* Sanity checks.... */
#if TICKS_PER_SLICE > 65500
#error Too many ticks. Reduce time slice, or increase prescaler
#elif TICKS_PER_SLICE == 0
#error Clock not running!
#elif TICKS_PER_SLICE < 100
#error Too less ticks. Increase time slice, or decrease prescaler
#endif

/* Prescaler mapper. */
#if PRESCALER == 1
#define PRESCALER_BITMAP 1
#elif PRESCALER == 8
#define PRESCALER_BITMAP 2
#elif PRESCALER == 64
#define PRESCALER_BITMAP 3
#elif PRESCALER == 256
#define PRESCALER_BITMAP 4
#elif PRESCALER == 1024
#define PRESCALER_BITMAP 5
#else
#error PRESCALER value is not supported!
#endif

/* We count too much ticks to the process, because saving */
/* and restoring regs takes time. We try to reduce that.  */
/* The values of 80 and 80 are found by examining the     */
/* assember code.                                         */
#define TICKS_IN_SAVE_CONTEXT ((uint8_t)(80/PRESCALER))
#define TICKS_IN_RESTORE_CONTEXT ((uint8_t)(80/PRESCALER))

/* Prototyping */
/* Naked attribute provides no register savings and return */
/* user must take care of it                               */
void TIMER1_COMPA_vect( void ) __attribute__ ( ( signal, naked ) );
/* Lost tasks handler */
void catch_lost_proc(void);
/* add a process */
void add_process(void *pfunc);
/* Clear the list */
void clear_process_list(void);
/* Voluntarily give up time slice */
void yield(void) __attribute__ ( ( naked ) );
/* Add the main routine */
void add_process_main(void);
/* Voluntarily give up time slice, semaphore version */
void sem_yield(void) __attribute__ ( ( naked ) );
/* Set up timer */
void setup_timer(void);


/* These routines are polity stolen from freeRTOS but modified */
/* for special need in this OS. First of all, register pushing */
/* and popping are reversed. Second, the flags member of the   */
/* process context is checked for BY_SEMFIELD (currently 4)    */
/* and return is a ret if so, otherwise a reti.                */
#define save_context() \
	asm volatile (	"; start with push regs.....	\n\t"	\
					"push	r31						\n\t"	\
					"in		r31, __SREG__			\n\t"	\
					"cli							\n\t"	\
					"push	r31	; flags				\n\t"	\
					"push	r30						\n\t"	\
					"push	r29						\n\t"	\
					"push	r28						\n\t"	\
					"push	r27						\n\t"	\
					"push	r26						\n\t"	\
					"push	r25						\n\t"	\
					"push	r24						\n\t"	\
					"push	r23						\n\t"	\
					"push	r22						\n\t"	\
					"push	r21						\n\t"	\
					"push	r20						\n\t"	\
					"push	r19						\n\t"	\
					"push	r18						\n\t"	\
					"push	r17						\n\t"	\
					"push	r16						\n\t"	\
					"push	r15						\n\t"	\
					"push	r14						\n\t"	\
					"push	r13						\n\t"	\
					"push	r12						\n\t"	\
					"push	r11						\n\t"	\
					"push	r10						\n\t"	\
					"push	r9						\n\t"	\
					"push	r8						\n\t"	\
					"push	r7						\n\t"	\
					"push	r6						\n\t"	\
					"push	r5						\n\t"	\
					"push	r4						\n\t"	\
					"push	r3						\n\t"	\
					"push	r2						\n\t"	\
					"push	r1						\n\t"	\
					"clr	r1						\n\t"	\
					"push	r0						\n\t"	\
					"; get pointer into X, so we can update the stackpointer... 	\n\t"	\
					"lds	r26, ptr				\n\t"	\
					"lds	r27, ptr + 1			\n\t"	\
					"in		r0, 0x3d				\n\t"	\
					"st		x+, r0					\n\t"	\
					"in		r0, 0x3e				\n\t"	\
					"st		x+, r0					\n\t"	\
				);

#define restore_context()									\
	asm volatile (	"; get pointer into X, so we can update the SP...	\n\t"	\
					"lds	r26, ptr				\n\t"	\
					"lds	r27, ptr + 1			\n\t"	\
					"ld		r28, x+					\n\t"	\
					"out	__SP_L__, r28			\n\t"	\
					"ld		r29, x+					\n\t"	\
					"out	__SP_H__, r29			\n\t"	\
					"pop	r0						\n\t"	\
					"pop	r1						\n\t"	\
					"pop	r2						\n\t"	\
					"pop	r3						\n\t"	\
					"pop	r4						\n\t"	\
					"pop	r5						\n\t"	\
					"pop	r6						\n\t"	\
					"pop	r7						\n\t"	\
					"pop	r8						\n\t"	\
					"pop	r9						\n\t"	\
					"pop	r10						\n\t"	\
					"pop	r11						\n\t"	\
					"pop	r12						\n\t"	\
					"pop	r13						\n\t"	\
					"pop	r14						\n\t"	\
					"pop	r15						\n\t"	\
					"pop	r16						\n\t"	\
					"pop	r17						\n\t"	\
					"pop	r18						\n\t"	\
					"pop	r19						\n\t"	\
					"pop	r20						\n\t"	\
					"pop	r21						\n\t"	\
					"pop	r22						\n\t"	\
					"pop	r23						\n\t"	\
					"pop	r24						\n\t"	\
					"pop	r25						\n\t"	\
					"pop	r26						\n\t"	\
					"pop	r27						\n\t"	\
					"; Now clear TC1 to zero, so	\n\t"	\
					"; the next process gets more	\n\t"	\
					"; or less 1 time slice.		\n\t"	\
					"clr	r28						\n\t"	\
					"out	%0, r28					\n\t"	\
					"out	%1, r28					\n\t"	\
					"; check pusher, via the flags	\n\t"	\
					"; member, if it is BY_SEMFIELD	\n\t"	\
					"; we do a ret (leaves I flag	\n\t"	\
					"; cleared and thus prohibits	\n\t"	\
					"; interrupts).					\n\t"	\
					"lds	r30, ptr				\n\t"	\
					"lds	r31, ptr + 1			\n\t"	\
					"adiw	r30, 2					\n\t"	\
					"ld		r29, z					\n\t"	\
					"andi	r29, %2					\n\t"	\
					"cpi	r29, %3					\n\t"	\
					"breq	L_dl2%=					\n\t"	\
					";								\n\t"	\
					"pop	r28						\n\t"	\
					"pop	r29						\n\t"	\
					"pop	r30						\n\t"	\
					"pop	r31						\n\t"	\
					"out	__SREG__, r31			\n\t"	\
					"pop	r31						\n\t"	\
					"reti							\n\t"	\
					";								\n\t"	\
		"L_dl2%=:	 pop	r28						\n\t"	\
					"pop	r29						\n\t"	\
					"pop	r30						\n\t"	\
					"pop	r31						\n\t"	\
					"out	__SREG__, r31			\n\t"	\
					"pop	r31						\n\t"	\
					"ret							\n\t"	\
					: : "I" (_SFR_IO_ADDR(TCNT1H)), "I" (_SFR_IO_ADDR(TCNT1L)), "I" (BY_MASK), "I" (BY_SEMYIELD)										\
			);

#endif
