/*

	semaphore.c  -  semaphore implementation

    Copyright (C) 2007  J. op den Brouw <jesse@opdenbrouw.nl>

	This file is part of jrtos version 0.2

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/


#include <avr/interrupt.h>
#include <avr/io.h>
#include <stdint.h>

#include "semaphore.h"

/* Initialization of semaphore */
void init_sem(semaphore *ps, uint8_t value) {
	*ps = value;
}

/* Up implementation, as is discussed in Operating Systems     */
/* Concepts, 2nd ed. by Peterson and Silberschatz, page 340.   */
/* The implementation is straight forward: it just increments  */
/* the semaphore. it doesn't do any wakeup on other processes. */
void up(semaphore *ps) {
	cli();
	(*ps)++;
#ifdef SEM_ROUTINE_TRICKY
	asm volatile ("reti");
#else
	/* Next is when not tricky */
	sei();
#endif
}

/* Down implementation, as is discussed in Operating Systems */
/* Concepts, 2nd ed. by Peterson and Silberschatz, page 340. */
/* The implementation is straight forward: it waits in a     */
/* while loop, testing the semaphore for <= 0, and when it   */
/* is, yielding the current process. This is very costly in  */
/* terms of CPU cycles, because we do a complete task switch.*/
void down(semaphore *ps) {
	cli();
	while (*ps<=0) {
		/* Please call sem_yield when yielding, because this */
		/* records the fact that on return, interrupts must  */
		/* be disabled (the while loop must be atomic).      */
		sem_yield();
	}
	(*ps)--;
#ifdef SEM_ROUTINE_TRICKY
	/* This is a very, VERY dirty hack and should be revisited */
	asm volatile ("pop r29");
	asm volatile ("pop r28");
	asm volatile ("reti");
#else
	/* From here on the 'normal' return procedure */
	sei();
#endif
	/* Here under you do not see anything in C, but watch the */
	/* assembler code: it pop r28 and r29 and does a ret.     */
}
